import { open, Database } from "sqlite";
import sqlite3 from "sqlite3";
import fs from "fs";
import _, { where } from "underscore";
import YAML from "yaml";
import {Card, CheckerInfo, loadCheckers } from "./src/checker";
import { AbstractChecker } from "./src/checkers/AbstractChecker";

type Where = string | any;

interface RegexCondition {
	pattern: string;
	field: 'name' | 'desc';
}

interface QueryInfo {
	upper: number;
	lower: number;
	noFilterAlias: boolean;
	where?: Where[];
	conditions?: CheckerInfo[];
	mayIncorrectStatements?: MayIncorrectStatement[];
}

interface MayIncorrectStatement {
	incorrectCountMin: number;
	incorrectCountMax: number;
	where: Where[];
}

let checkersList: Map<string, AbstractChecker>;
let db: Database;
let constantDict = new Map<string, number>();
constantDict.set("TYPES_EXTRA", 0x20 | 0x4000 | 0x800000 | 0x4000000)

async function loadConstants() {
	try {
		const commonContent = await fs.promises.readFile("./ocgcore/common.h", "utf-8");
		for (let lineRaw of commonContent.split("\n")) {
			const line = lineRaw.trim();
			const match = line.match(/^#define ([_A-Z0-9]+)\t+(0x[0-9a-fA-F]+)/);
			if (match) {
				constantDict.set(match[1], parseInt(match[2]));
			}
		}
	} catch (e) {
		console.error(`Load dict failed: ${e.toString()}`);
	}

}

function joinWithAnd(clauses: string[]) {
	return clauses.join(" ");
}
function joinWithOr(clauses: string[]) {
	const wrappedClauses = clauses.map(c => c.slice(4));
	return `and (${wrappedClauses.join(" or ")})`;
}
function wrapWithNot(clause: string) {
	return `and not ${clause.slice(4)}`;
}

function replaceClause(clause: string | any, statement: boolean) {
	if (typeof (clause) !== "string") {
		return (statement === (clause === 0)) ? "and 0" : "and 1";
	}
	let newClause = clause;
	for (let constantKey of Array.from(constantDict.keys())) {
		const regex = new RegExp(`\{${constantKey}\}`, "g");
		newClause = newClause.replace(regex, constantDict.get(constantKey).toString());
	}
	if (statement) {
		return `and (${newClause})`;
	} else {
		return `and not (${newClause})`;
	}
}

function parseWhereClause(whereClause: (string | any)[]) {
	if (!whereClause || !whereClause.length) {
		return "";
	}
	const newClauses = whereClause.map(w => replaceClause(w, true));
	return joinWithAnd(newClauses);
}

function getCombination(arr: number[], count: number): number[][] {
	const len = arr.length;
	if (count >= len || count === 0) {
		return [arr.slice(0, count)]
	} else if (count > len / 2) {
		const reversedCombination = getCombination(arr, len - count);
		return reversedCombination.map(m => _.without(arr, ...m))
	} else {
		const [first, ...other] = arr;
		let res: number[][] = [];
		const [combinationsWithFirst, combinationsOthers]: [number[][], number[][]] = [
			getCombination(other, count - 1),
			getCombination(other, count)
		];
		res = res.concat(combinationsWithFirst.map(comb => [first, ...comb]))
		res = res.concat(combinationsOthers);
		return res;
	}

}

function generateBoolSequence(totalCount: number, falseCount: number): boolean[][] {
	const allTrueSequence = _.range(totalCount).map(m => true);
	const falsePositions = getCombination(_.range(totalCount), falseCount);
	return falsePositions.map(falsePosition => {
		const currentSequence = _.clone(allTrueSequence);
		for (let pos of falsePosition) {
			currentSequence[pos] = false;
		}
		return currentSequence;
	})
}

function parseMayIncorrectStatement(mayIncorrectStatement: MayIncorrectStatement) {
	if (!mayIncorrectStatement.where.length) {
		return "";
	}
	let outerClauses: string[] = [];
	for (let i = mayIncorrectStatement.incorrectCountMin; i <= mayIncorrectStatement.incorrectCountMax; ++i) {
		const correctSequences = generateBoolSequence(mayIncorrectStatement.where.length, i);
		for (let correctSequence of correctSequences) {
			const clauses = mayIncorrectStatement.where.map((clause, index) => replaceClause(clause, correctSequence[index]));
			outerClauses.push(joinWithAnd(clauses));
		}
	}
	return joinWithOr(outerClauses);
}

function parseMayIncorrectStatements(mayIncorrectStatements: MayIncorrectStatement[]) {
	if (!mayIncorrectStatements || !mayIncorrectStatements.length) {
		return "";
	}
	return mayIncorrectStatements.map(m => parseMayIncorrectStatement(m)).join(" ");
}

async function openDatabase() {
	return await open({
		filename: "./cards.cdb",
		driver: sqlite3.Database
	});
}

function getMedianCode(cards: Card[]) {
	const sortedCards = cards;
	const len = sortedCards.length;
	if (len % 2 === 0) {
		return Math.floor((sortedCards[len / 2 - 1].id + sortedCards[len / 2].id) / 2);
	} else {
		const targetSequence = (len + 1) / 2 - 1;
		return Math.floor((sortedCards[targetSequence - 1].id + sortedCards[targetSequence].id) / 2);
	}
}

async function queryCards(queryInfo: QueryInfo) {
	const whereClauseString = parseWhereClause(queryInfo.where);
	const mayIncorrectClauseString = parseMayIncorrectStatements(queryInfo.mayIncorrectStatements);
	const sql = `select datas.id,texts.name,texts.desc from datas,texts where datas.id = texts.id and datas.type & 0x4000 = 0 and datas.id > ? and datas.id <= ? ${queryInfo.noFilterAlias ? "" : "and (datas.alias = 0 or datas.id - datas.alias > 10)"} ${whereClauseString} ${mayIncorrectClauseString} order by datas.id asc`;
	console.error(`SQL: ${sql}`);
	let cards: Card[] = await db.all(sql, [queryInfo.lower, queryInfo.upper]);
	if (queryInfo.conditions) {
		for (let checkerInfo of queryInfo.conditions) {
			const checker = checkersList.get(checkerInfo.type);
			if (!checker) {
				console.error(`Checker ${checkerInfo.type} not found.`);
				continue;
			}
			const wantedStatement = checkerInfo.statement != null ? checkerInfo.statement : true;
			await checker.setInfo(checkerInfo.data);
			cards = cards.filter(card => !!checker.check(card) === wantedStatement);
		}
	}
	return cards;
}

async function loadYAML(path: string) {
	const content = await fs.promises.readFile(path, "utf-8");
	return YAML.parse(content);
}

async function main() {
	checkersList = await loadCheckers();
	db = await openDatabase();
	const queryInfo: QueryInfo = await loadYAML("./info.yaml");
	await loadConstants();
	const cards = await queryCards(queryInfo);
	console.log(`Cards:`);
	for (let card of cards) {
		console.log(`${card.id} ${card.name}`);
	}
	console.log(`Size: ${cards.length}`);
	if (cards.length > 1) {
		const median = getMedianCode(cards);
		console.log(`Median: ${median}`);
	}
}
main()
